import os
from argparse import Namespace
from unittest import TestCase
from unittest.mock import patch

from shub_workflow.script import BaseScript, BaseScriptProtocol
from shub_workflow.base import WorkFlowManager, WorkFlowManagerProtocol
from shub_workflow.graph import GraphManager
from shub_workflow.graph.task import Task
from shub_workflow.crawl import GeneratorCrawlManager, AsyncSchedulerCrawlManagerMixin
from shub_workflow.deliver import BaseDeliverScript
from shub_workflow.utils.contexts import script_args


class MyDeliverScript(BaseDeliverScript):
    SCRAPERNAME_NARGS = 1


class MyMixin(BaseScriptProtocol):
    def __init__(self):
        super().__init__()
        self.target = 0

    def append_flow_tag(self, tag: str):
        self.target = 1

    def parse_args(self) -> Namespace:
        args = super().parse_args()  # type: ignore
        if hasattr(args, "exec_id"):
            self.append_flow_tag(f"EXEC_ID={args.exec_id}")
        return args


class MyWFMixin(WorkFlowManagerProtocol):
    def __init__(self):
        super().__init__()
        self.target = 0

    def append_flow_tag(self, tag: str):
        self.target = 1


class MyScript(MyMixin, BaseScript):
    def run(self):
        pass


class MyWorkFlowManager(MyMixin, WorkFlowManager):
    def workflow_loop(self) -> bool:
        return False


class MyWorkFlowManagerTwo(MyWFMixin, WorkFlowManager):
    def workflow_loop(self) -> bool:
        return False


class MyGraphManager(MyMixin, GraphManager):
    def configure_workflow(self):
        jobA = Task("jobA", "py:command.py")
        return (jobA,)


class MyAsyncCrawlManager(AsyncSchedulerCrawlManagerMixin, GeneratorCrawlManager):  # type: ignore
    def set_parameters_gen(self):
        yield from ()


@patch("shub_workflow.script.BaseScript.get_sc_project_settings", new=lambda _: {})
class TypingTest(TestCase):
    def setUp(self):
        os.environ["SH_APIKEY"] = "ffff"
        os.environ["PROJECT_ID"] = "999"

    @patch("shub_workflow.script.BaseScript.add_job_tags")
    def test_script_instantiation(self, mocked_add_job_tags):
        with script_args([]):
            script = MyScript()
            script.append_flow_tag("mytag")
        # self.assertEqual(mocked_add_job_tags.call_count, 0)
        self.assertEqual(script.target, 1)

    @patch("shub_workflow.script.BaseScript.add_job_tags")
    def test_workflow_manager_instantiation(self, mocked_add_job_tags):
        with script_args(["myname"]):
            manager = MyWorkFlowManager()
            manager.append_flow_tag("mytag")
        # self.assertEqual(mocked_add_job_tags.call_count, 0)
        self.assertEqual(manager.target, 1)

    @patch("shub_workflow.script.BaseScript.add_job_tags")
    def test_workflow_manager_instantiation_two(self, mocked_add_job_tags):
        with script_args(["myname"]):
            manager = MyWorkFlowManagerTwo()
            manager.append_flow_tag("mytag")
        # self.assertEqual(mocked_add_job_tags.call_count, 0)
        self.assertEqual(manager.target, 1)

    @patch("shub_workflow.script.BaseScript.add_job_tags")
    def test_graph_manager_instantiation(self, mocked_add_job_tags):
        with script_args(["myname"]):
            manager = MyGraphManager()
            manager.append_flow_tag("mytag")
        # self.assertEqual(mocked_add_job_tags.call_count, 0)
        self.assertEqual(manager.target, 1)

    def test_asyncrawlmanager(self):
        with script_args(["myname", "myspider"]):
            MyAsyncCrawlManager()



import logging
from pprint import pformat

from scrapy.signals import item_scraped

from shub_workflow.script import BaseScriptProtocol


LOGGER = logging.getLogger(__name__)


class ItemHSIssuerMixin(BaseScriptProtocol):
    """
    A class for allowing to issue items on hubstorage, so a script running on SC can return items as a spider.
    """

    def __init__(self):
        super().__init__()
        try:
            from sh_scrapy.extension import HubstorageExtension

            class _HubstorageExtension(HubstorageExtension):
                def item_scraped(slf, item, spider):
                    try:
                        return super().item_scraped(item, spider)
                    except RuntimeError:
                        self.log_item(item, spider)

            self.hextension = _HubstorageExtension.from_crawler(self._pseudo_crawler)
        except ImportError:
            self._pseudo_crawler.signals.connect(self.log_item, item_scraped)

    def log_item(self, item, spider, **kwargs):
        LOGGER.info(pformat(item))

    def hs_issue_item(self, item):
        self._pseudo_crawler.signals.send_catch_log_deferred(item_scraped, dont_log=True, item=item, spider=self)



import logging
import json
import os
import hashlib
from typing import Optional

from scrapy.settings import BaseSettings
from tenacity import retry, retry_if_exception_type, before_sleep_log, stop_after_attempt, wait_fixed
from scrapinghub.client.exceptions import ServerError
from requests.exceptions import ReadTimeout, ConnectionError, HTTPError


logger = logging.getLogger(__name__)


def hashstr(text: str) -> str:
    u = hashlib.sha1()
    u.update(text.encode("utf8"))
    return u.hexdigest()


def resolve_shub_jobkey() -> Optional[str]:
    return os.environ.get("SHUB_JOBKEY")


def resolve_project_id(project_id=None) -> int:
    """
    Gets project id from following sources in following order of precedence:
    - default parameter values
    - environment variables
    - scrapinghub.yml file

    in order to allow to use codes that needs HS or dash API,
    either locally or from scrapinghub, correctly configured
    """
    if project_id:
        try:
            retval = int(project_id)
            logger.info(f"Project id {retval} is explicit.")
            return retval
        except ValueError:
            pass
    else:
        # read from environment only if not explicitly provided
        if os.environ.get("PROJECT_ID") is not None:
            retval = int(os.environ["PROJECT_ID"])
            logger.info(f"Project id {retval} resolved from PROJECT_ID environment variable.")
            return retval

        # for ScrapyCloud jobs:
        jobkey = resolve_shub_jobkey()
        if jobkey:
            retval = int(jobkey.split("/")[0])
            logger.info(f"Project id {retval} resolved from jobkey.")
            return retval

    # read from scrapinghub.yml
    try:
        from shub.config import load_shub_config

        cfg = load_shub_config()
        try:
            project_id = project_id or "default"
            retval = int(cfg.get_project_id(project_id))
            logger.info(f"Project id {retval} resolved from scrapinghub.yml {project_id} entry.")
            return retval
        except Exception:
            logger.error(f"Project entry '{project_id}' not found in scrapinghub.yml.")
    except ImportError:
        logger.error("Install shub package if want to access scrapinghub.yml.")
    except TypeError:
        logger.error("Default project entry not available in scrapinghub.yml.")

    raise ValueError(
        "No default project id found. Use either PROJECT_ID env. variable or set 'default' entry in scrapinghub.yml, "
        "or use --project-id with a project numeric id or an existing entry in scrapinghub.yml."
    )


MINS_IN_A_DAY = 24 * 60
ONE_MIN_IN_S = 60


DASH_RETRY_MAX = int(os.environ.get("DASH_RETRY_MAX", MINS_IN_A_DAY))
DASH_RETRY_WAIT_SECS = int(os.environ.get("DASH_RETRY_WAIT_SECS", ONE_MIN_IN_S))
DASH_RETRY_LOGGING_LEVEL = os.environ.get("DASH_RETRY_LOGGING_LEVEL", "ERROR")


dash_retry_decorator = retry(
    # ServerError is the only ScrapinghubAPIError that should be retried. Don't capture ScrapinghubAPIError here
    retry=retry_if_exception_type((ServerError, ReadTimeout, ConnectionError, HTTPError)),
    before_sleep=before_sleep_log(logger, getattr(logging, DASH_RETRY_LOGGING_LEVEL)),
    reraise=True,
    stop=stop_after_attempt(DASH_RETRY_MAX),
    wait=wait_fixed(DASH_RETRY_WAIT_SECS),
)

_settings_warning_issued = False


def kumo_settings():
    global _settings_warning_issued
    settings = {}
    shub_job_data = json.loads(os.environ.get("SHUB_SETTINGS", "{}"))
    if shub_job_data:
        settings.update(shub_job_data["project_settings"])
        settings.update(shub_job_data["spider_settings"])
    elif not _settings_warning_issued:
        logger.warning("Couldn't find Dash project settings, probably not running in Dash")
        _settings_warning_issued = True
    return settings


def get_project_settings() -> BaseSettings:
    from scrapy.utils.project import get_project_settings as scrapy_get_project_settings  # pylint: disable=import-error

    settings = scrapy_get_project_settings()
    settings.setdict(kumo_settings(), priority="project")
    try:
        # test sh_scrapy is available
        import sh_scrapy  # noqa: F401

        settings.setdict({"STATS_CLASS": "sh_scrapy.stats.HubStorageStatsCollector"}, priority="cmdline")
    except ImportError:
        pass
    return settings


def get_kumo_loglevel(default="INFO"):
    loglevel = kumo_settings().get("LOG_LEVEL", default)
    return getattr(logging, loglevel)



import re
import abc
import time
import logging
import inspect
import csv
from io import StringIO
from typing import Dict, Type, Tuple, Optional, Protocol, Union
from datetime import timedelta, datetime
from collections import Counter

import dateparser
from scrapy import Spider
from prettytable import PrettyTable

from shub_workflow.script import BaseScript, BaseScriptProtocol, SpiderName, JobDict
from shub_workflow.utils.alert_sender import AlertSenderMixin
from shub_workflow.contrib.slack import SlackSender

LOG = logging.getLogger(__name__)


def _get_number(txt: str) -> Optional[int]:
    try:
        return int(txt)
    except ValueError:
        return None


class BaseMonitorProtocol(BaseScriptProtocol, Protocol):

    @abc.abstractmethod
    def close(self):
        pass


BASE_TARGET_SPIDER_STATS = (
    "downloader/response_status_count/",
    "downloader/response_count",
    "item_scraped_count",
    "spider_exceptions/",
    "scrapy-zyte-api/429",
    "zyte_api_proxy/response/status/429",
)


RESPONSE_STATUS_COUNT_RE = re.compile(r"downloader/response_status_count/(\d+)/(.+)")


class SpiderStatsAggregatorMixin(BaseScriptProtocol):
    # stats aggregated from spiders. A tuple of stats regex prefixes.
    target_spider_stats: Tuple[str, ...] = ()
    stats_only_total = True

    def aggregate_spider_stats(self, jobdict: JobDict, stats_added_prefix: str = ""):
        canonical = self.get_canonical_spidername(jobdict["spider"])
        for statkey in jobdict.get("scrapystats") or {}:
            for statnameprefix in self.target_spider_stats + BASE_TARGET_SPIDER_STATS:
                if re.match(statnameprefix, statkey) is not None:
                    value = jobdict["scrapystats"][statkey]
                    if stats_added_prefix != canonical:
                        if not self.stats_only_total:
                            self.stats.inc_value(f"{stats_added_prefix}/{statkey}/{canonical}".strip("/"), value)
                        self.stats.inc_value(f"{stats_added_prefix}/{statkey}/total".strip("/"), value)
                    else:
                        self.stats.inc_value(f"{stats_added_prefix}/{statkey}".strip("/"), value)


class BaseMonitor(AlertSenderMixin, SpiderStatsAggregatorMixin, BaseScript, BaseMonitorProtocol):

    # a map from spiders classes to check, to a stats prefix to identify the aggregated stats.
    target_spider_classes: Dict[Type[Spider], str] = {Spider: ""}

    # - a map from script name into a tuple of 2-elem tuples (aggregating stat regex, aggregated stat prefix)
    # - the aggregating stat regex is used to match stat on target script
    # - the aggregated stat prefix is used to generate the monitor stat. The original stat name is appended to
    #   the prefix.
    # - if a group is present in the regex, its value is used as suffix of the generate stat, instead of
    #   the complete original stat name.
    target_script_stats: Dict[str, Tuple[Tuple[str, str], ...]] = {}

    # - a map from script name into a tuple of 2-elem tuples (aggregating log regex, aggregated stat name)
    # - the aggregating log regex must match log lines in the target script job, with a first group to extract a number
    #   from it. If not a group number is extracted, the match alone aggregates 1.
    # - the aggregated stat name is the name where to aggregate the number extracted by the regex, plus a second group,
    #   if exists.

    # - a map from script name into a tuple of 2-elem tuples (aggregating log regex, aggregated stat name)
    # - the aggregating log regex must match log lines in the target script job, with a group to extract a number
    #   from it. If not a group number is extracted, the match alone aggregates 1.
    # - the final aggregated stat name is the specified aggregated stat name, plus a second non numeric group in the
    #   match, if exists.
    target_script_logs: Dict[str, Tuple[Tuple[str, str], ...]] = {}

    # Define here ratios computations. Each 3-tuple contains:
    # - a string or regex that matches the numerator stat. You can use a regex group here for computing multiple ratios
    # - a string or regex that matches the denominator stat. You can use a regex group for matching numerator group and
    #   have a different denominator for each one. If no regex group in denominator, there will be a single
    #   denominator.
    # - the target stat. If there are numerator groups, this will be the target stat prefix.
    # Ratios computing are performed after the stats_postprocessing() method is called. It is a stat postprocessing
    #   itself.
    stats_ratios: Tuple[Tuple[str, str, str], ...] = ()

    # A tuple of 2-elem tuples each one with a stat regex and the name of the monitor instance method that will receive:
    # - start and end limit of the window (epoch seconds)
    # - the value of the stat
    # - one extra argument per regex group, if any.
    # Useful for adding monitor alerts or any kind of reaction according to stat value.
    stats_hooks: Tuple[Tuple[str, str], ...] = ()

    # if True, only generate the totals and not per crawler stats. This has effect on default stats only, not the
    # custom ones added by the developer.
    stats_only_total = False

    # A tuple of string tuples for generating a report table.
    # The first line is the header. Following ones are the rows. Ensure that all tuples has the same length.
    report_table: Tuple[Tuple[str, ...], ...] = ()

    # additional projects in multiproject projects
    additional_projects: Tuple[int, ...] = ()

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument(
            "--period", "-p", type=int, default=86400, help="Time window period in seconds. Default: %(default)s"
        )
        self.argparser.add_argument(
            "--end-time",
            "-e",
            type=str,
            help="""End side of the time window. By default it is just now. Format: any string
that can be recognized by dateparser.""",
        )
        self.argparser.add_argument(
            "--start-time",
            "-s",
            type=str,
            help="""
            Starting side of the time window. By default it is the end side minus the period. Format:
            any string that can be recognized by dateparser.
        """,
        )
        self.argparser.add_argument(
            "--generate-report",
            action="store_true",
            help=("Generate report table. See report_table attribute and generate_report() method."),
        )
        self.argparser.add_argument(
            "--report-format",
            choices=("csv", "pretty", "pretty_with_borders", "pretty_with_tabs"),
            default="pretty",
            help="'pretty_with_tabs' is suitable for easy copy and paste over a spreadsheet.",
        )
        self.argparser.add_argument(
            "--slack-report", action="store_true", help="Send report to slack. By default it just prints it in the log."
        )

    def stats_postprocessing(self, start_limit, end_limit):
        """
        Apply here any additional code for post processing stats, generate derivated stats, reports, etc.
        """

    def generate_report(self):
        """
        Facilitates the generation of printed/slack reports.
        """
        header: Tuple[str, ...] = self.report_table[0]
        rows: Tuple[Tuple[str, ...], ...] = self.report_table[1:]

        table = PrettyTable(
            field_names=header,
            border=self.args.report_format == "pretty_with_borders",
            preserve_internal_border=self.args.report_format == "pretty_with_tabs",
        )
        for row in rows:
            table.add_row(list(row))

        if self.args.report_format == "csv":
            fp = StringIO()
            w = csv.writer(fp)
            w.writerow(table.field_names)
            w.writerows(table.rows)
            fp.seek(0)
            table_text = fp.read()
            fp.close()
        elif self.args.report_format == "pretty_with_tabs":
            table_text = str(table).replace(" | ", " \t ")
        else:
            table_text = str(table)

        if self.args.slack_report:
            table_text = "```" + table_text + "```"
            sender = SlackSender(self.project_settings)
            sender.send_slack_messages(
                [table_text],
                self.args.subject or "Stats Report",
            )
            # avoid slack rate limit error when sending the rates alert to slack
            time.sleep(30)
        else:
            print(table_text)

    def run(self):
        end_limit = time.time()
        if self.args.end_time is not None and (dt := dateparser.parse(self.args.end_time)) is not None:
            end_limit = dt.timestamp()
        start_limit = end_limit - self.args.period
        if self.args.start_time and (dt := dateparser.parse(self.args.start_time)) is not None:
            start_limit = dt.timestamp()
        else:
            LOG.info(f"Period set: {timedelta(seconds=self.args.period)}")
        LOG.info(f"Start time: {str(datetime.fromtimestamp(start_limit))}")
        LOG.info(f"End time: {str(datetime.fromtimestamp(end_limit))}")

        for attr in dir(self):
            if attr.startswith("check_"):
                method = getattr(self, attr)
                if inspect.ismethod(method):
                    check_name = attr.replace("check_", "")
                    LOG.info(f"Checking {check_name}...")
                    method(start_limit, end_limit)

        self.stats_postprocessing(start_limit, end_limit)
        self.run_stats_ratios()
        if self.args.generate_report:
            self.generate_report()
        self.run_stats_hooks(start_limit, end_limit)
        self.upload_stats()
        self.print_stats()
        self.close()

    def run_stats_ratios(self):
        """
        Generate new ratio stats, based on definitions in stats_ratios attribute.
        """

        stats_ratios = self.stats_ratios

        # append response status ratios
        for prefix in self.target_spider_classes.values():
            if prefix:
                prefix = prefix + "/"
            for stat, value in list(self.stats.get_stats().items()):
                if (m := RESPONSE_STATUS_COUNT_RE.search(stat)) is not None and stat.startswith(prefix):
                    status_code, spider = m.groups()
                    stats_ratios += ((
                        f"{prefix}downloader/response_status_count/{status_code}/{spider}",
                        f"{prefix}downloader/response_count/{spider}",
                        f"{prefix}downloader/response_count/rate/{status_code}/{spider}"
                    ),)

        for num_regex, den_regex, target_prefix in stats_ratios:
            numerators: Dict[str, int] = Counter()
            denominators: Dict[str, int] = Counter()
            numerator_has_groups = False
            denominator_has_groups = False
            for stat, value in list(self.stats.get_stats().items()):
                if m := re.search(num_regex, stat):
                    if m.groups():
                        source = m.groups()[0]
                        numerators[source] += value
                        numerator_has_groups = True
                    else:
                        numerators[stat] += value
                if m := re.search(den_regex, stat):
                    if m.groups():
                        source = m.groups()[0]
                        denominators[source] += value
                        denominator_has_groups = True
                    else:
                        denominators[stat] += value
            for source, numer in numerators.items():
                denominator = 0
                if denominator_has_groups:
                    denominator = denominators.pop(source, 0)
                elif denominators:
                    denominator = list(denominators.values())[0]
                    denominators = {}
                if denominator > 0:
                    target_stat = target_prefix
                    if numerator_has_groups:
                        target_stat += "/" + source
                    self.stats.set_value(target_stat, round(numer / denominator, 4))
            for source, denom in denominators.items():
                target_stat = target_prefix
                if numerator_has_groups:
                    target_stat += "/" + source
                self.stats.set_value(target_stat, round(0, 4))

    def run_stats_hooks(self, start_limit, end_limit):
        for stat, val in self.stats.get_stats().items():
            for stat_re, hook_name in self.stats_hooks:
                if (m := re.search(stat_re, stat)) is not None:
                    hook = getattr(self, hook_name)
                    hook(start_limit, end_limit, self.stats.get_value(stat), *m.groups())

    def _get_stats_prefix_from_spider_class(self, spiderclass: Type[Spider]) -> str:
        for cls, prefix in self.target_spider_classes.items():
            if issubclass(spiderclass, cls):
                return prefix
        return ""

    def spider_job_hook(self, jobdict: JobDict):
        """
        This is called for every spider job retrieved, from the spiders declared on target_spider_classes,
        so additional per job customization can be added.
        """

    def get_jobs_in_window(self, start_limit: int, end_limit: Union[int, float, None], **kwargs):
        end_limit = end_limit or float("inf")
        for project_id in (self.project_id,) + self.additional_projects:
            for jobdict in self.get_jobs(project_id=project_id, **kwargs):
                if "finished_time" in jobdict and jobdict["finished_time"] / 1000 < start_limit:
                    break
                if "finished_time" in jobdict and jobdict["finished_time"] / 1000 > end_limit:
                    continue
                yield jobdict

    def check_spiders(self, start_limit, end_limit):

        spiders: Dict[SpiderName, Type[Spider]] = {}
        for s in self.spider_loader.list():
            subclass = self.spider_loader.load(s)
            if issubclass(subclass, tuple(self.target_spider_classes.keys())):
                spiders[SpiderName(s)] = subclass

        for jobcount, jobdict in enumerate(
            self.get_jobs_in_window(
                start_limit,
                end_limit,
                state=["finished"],
                meta=["spider", "finished_time", "scrapystats", "spider_args", "close_reason", "tags"],
                has_tag=[f"FLOW_ID={self.flow_id}"] if self.flow_id is not None else None,
                endts=int(end_limit * 1000),
            ),
            start=1,
        ):
            if jobdict["spider"] in spiders:
                self.spider_job_hook(jobdict)
                stats_added_prefix = self._get_stats_prefix_from_spider_class(spiders[jobdict["spider"]])
                self.aggregate_spider_stats(jobdict, stats_added_prefix)
            if jobcount % 1000 == 0:
                LOG.info(f"Read {jobcount} jobs")

    def script_job_hook(self, jobdict: JobDict):
        """
        This is called for every script job retrieved, from the scripts declared on target_script_stats,
        so additional per job customization can be added
        """

    def check_scripts_stats(self, start_limit, end_limit):
        for script, regexes in self.target_script_stats.items():
            plural = script.replace("py:", "").replace(".py", "") + "s"
            LOG.info(f"Checking {plural} stats ...")
            for jobdict in self.get_jobs_in_window(
                start_limit,
                end_limit,
                spider=script,
                state=["finished"],
                meta=["finished_time", "scrapystats", "close_reason", "job_cmd", "tags"],
                endts=int(end_limit * 1000),
                has_tag=[f"FLOW_ID={self.flow_id}"] if self.flow_id is not None else None,
            ):
                self.script_job_hook(jobdict)
                for key, val in jobdict.get("scrapystats", {}).items():
                    for regex, prefix in regexes:
                        if (m := re.search(regex, key)) is not None:
                            aggregated_stat_name = m.groups()[0] if m.groups() else m.group()
                            self.stats.inc_value(f"{prefix}/{aggregated_stat_name}".strip("/"), val)
                            if prefix:
                                self.stats.inc_value(f"{prefix}/total".strip("/"), val)

    def check_script_logs(self, start_limit, end_limit):
        for script, regexes in self.target_script_logs.items():
            plural = script.replace("py:", "").replace(".py", "") + "s"
            LOG.info(f"Checking {plural} logs ...")
            for jobdict in self.get_jobs_in_window(
                start_limit,
                None,
                spider=script,
                state=["running", "finished"],
                meta=["state", "finished_time"],
                has_tag=[f"FLOW_ID={self.flow_id}"] if self.flow_id is not None else None,
            ):
                if jobdict["state"] == "running" or jobdict["finished_time"] / 1000 > start_limit:
                    job = self.get_job(jobdict["key"])
                    for logline in job.logs.iter():
                        if logline["time"] / 1000 < start_limit:
                            continue
                        if logline["time"] / 1000 > end_limit:
                            break
                        for regex, stat in regexes:
                            if (m := re.search(regex, logline["message"])) is not None:
                                stat_suffix = ""
                                if gr := m.groups():
                                    val = _get_number(gr[0])
                                    if val is None:
                                        if len(gr) > 1:
                                            val = _get_number(gr[1])
                                            stat_suffix = gr[0]
                                        if val is None:
                                            val = 1
                                    else:
                                        stat_suffix = gr[1] if len(gr) > 0 else ""
                                else:
                                    val = 1
                                self.stats.inc_value(stat, val)
                                if stat_suffix:
                                    self.stats.inc_value(stat + f"/{stat_suffix}", val)

    def close(self):
        self.send_messages()



"""
Watchdog script.

It checks for failed scripts, and send alerts in case issues are detected. Optionally it can clone
failed scripts, typically workflow managers and other standalone scripts. Scripts handled by workflow managers
should not be retried here.

Spiders are handled by crawl managers, so they are not handled here.

Script configurable attributes:

MONITORED_SCRIPTS - A tuple containing all scripts to check (each one in the format "py:scriptname.py")
CLONE_SCRIPTS - A tuple containing all scripts that must be cloned in case of failed. This tuple must be
                a subset of MONITORED_SCRIPTS.


"""

import time
import abc
import logging
from typing import List, Tuple

from shub_workflow.script import JobKey, Outcome, JobDict
from shub_workflow.utils.clone_job import BaseClonner


WATCHDOG_CHECKED_TAG = "WATCHDOG_CHECKED=True"
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class WatchdogBaseScript(BaseClonner):

    MONITORED_SCRIPTS: Tuple[str, ...] = ()
    CLONE_SCRIPTS: Tuple[str, ...] = ()
    DEFAULT_SPIDER_MAX_RUNNING_TIME_SECS: int = 3600 * 24 * 365
    CHECK_RUNNING_SPIDERS = False

    def __init__(self) -> None:
        super().__init__()
        self.failed_jobs: List[Tuple[str, JobKey, Outcome]] = []
        self.__notification_lines: List[str] = []

    def add_argparser_options(self):
        super().add_argparser_options()
        self.argparser.add_argument("period", help="How much go to past to check (in hours)", type=int)

    def append_notification_line(self, line: str):
        self.__notification_lines.append(line)

    def get_notification_lines(self) -> List[str]:
        return self.__notification_lines

    def run(self) -> None:
        self.check_failed_scripts()
        if self.CHECK_RUNNING_SPIDERS:
            self.check_running_spider_jobs()
        self.close()

    def check_failed_scripts(self) -> None:
        now = time.time()
        for script in self.MONITORED_SCRIPTS:
            count = 0
            for job in self.get_jobs(
                spider=script,
                state=["finished"],
                meta=["finished_time", "close_reason"],
                lacks_tag=WATCHDOG_CHECKED_TAG,
            ):
                age = (now - job["finished_time"] / 1000) / 3600
                if age > self.args.period:
                    break
                count += 1
                if job["close_reason"] != "finished":
                    msg = (
                        f"Failed task: {script} (job: https://app.zyte.com/p/{job['key']}) "
                        f"(Reason: {job['close_reason']})"
                    )
                    LOGGER.error(msg)
                    new_job = None
                    if script in self.CLONE_SCRIPTS:
                        new_job = self.clone_job(job["key"])
                    if new_job is None:
                        self.append_notification_line(msg)
                self.add_job_tags(job["key"], tags=[WATCHDOG_CHECKED_TAG])
            LOGGER.info(f"Checked {count} {script} jobs")

    def get_spider_job_max_running_time(self, job: JobDict) -> int:
        """
        Return max running job time in seconds
        """
        return self.DEFAULT_SPIDER_MAX_RUNNING_TIME_SECS

    def check_running_spider_jobs(self) -> None:
        now = time.time()
        for job in self.get_jobs(
            state=["running"], meta=["spider", "spider_args", "running_time", "job_settings"]
        ):
            if job["spider"].startswith("py:"):
                continue
            max_seconds = self.get_spider_job_max_running_time(job)
            if max_seconds:
                running_time = now - job["running_time"] / 1000
                if running_time > int(max_seconds) * 1.1:
                    self.finish(job["key"], close_reason="cancelled (watchdog)")
                    msg = f"Cancelled job https://app.zyte.com/p/{job['key']} (running for {int(running_time)} seconds)"
                    LOGGER.warning(msg)
                    self.append_notification_line(msg)

    def close(self) -> None:
        if self.__notification_lines:
            self.send_alert()

    @abc.abstractmethod
    def send_alert(self):
        ...



"""
Base script class for spiders crawl managers.
"""
import abc
import json
import asyncio
import logging
from copy import deepcopy
from argparse import Namespace
from typing import Optional, List, Tuple, Dict, NewType, cast, Generator, Any, AsyncGenerator, Set, Protocol, TypedDict

from bloom_filter2 import BloomFilter

from shub_workflow.script import (
    JobKey,
    JobDict,
    JobMeta,
    SpiderName,
    Outcome,
    BaseLoopScriptAsyncMixin,
)
from shub_workflow.base import WorkFlowManager
from shub_workflow.utils import hashstr
from shub_workflow.utils.dupefilter import DupesFilterProtocol
from shub_workflow.utils.monitor import SpiderStatsAggregatorMixin


_LOG = logging.getLogger(__name__)


SpiderArgs = NewType("SpiderArgs", Dict[str, str])
ScheduleArgs = NewType("ScheduleArgs", Dict[str, Any])


class JobParams(TypedDict, total=False):
    units: int
    tags: List[str]
    job_settings: Dict[str, str]
    project_id: int
    spider_args: Dict[str, str]


class FullJobParams(JobParams, total=False):
    spider: SpiderName


class StopRetry(Exception):
    pass


def get_jobseq(tags: List[str]) -> Tuple[int, int]:
    """
    returns tuple (job sequence, retry index)
    """
    for tag in tags:
        if tag.startswith("JOBSEQ="):
            tag = tag.replace("JOBSEQ=", "")
            jobseq, *rep = tag.split(".r")
            rep.append("0")
            return int(jobseq), int(rep[0])
    return 0, 0


class CrawlManagerProtocol(Protocol):
    _running_job_keys: Dict[JobKey, Tuple[SpiderName, JobParams]]

    @abc.abstractmethod
    def check_running_jobs(self) -> Dict[JobKey, Outcome]:
        ...

    @abc.abstractmethod
    def schedule_spider_with_jobargs(
        self,
        job_args_override: Optional[JobParams] = None,
        spider: Optional[SpiderName] = None,
    ) -> Optional[JobKey]:
        ...

    @abc.abstractmethod
    def get_job_settings(self, override: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        ...


class CrawlManager(SpiderStatsAggregatorMixin, WorkFlowManager, CrawlManagerProtocol):
    """
    Schedules a single spider job. If loop mode is enabled, it will shutdown only after the scheduled spider
    finished. Close reason of the manager will be inherited from spider one.
    """

    spider: Optional[SpiderName] = None

    # If False, check running jobs from older to newer
    # If True, check running jobs from newer to older.
    # The check order may have some implications on some applications
    # (for example, when analyzing running jobs for implementing
    # throttling algorithms based on concurrency, we may want always
    # the data from the most recent job)
    running_jobs_reverse_check = False

    def __init__(self):
        super().__init__()

        # running jobs represents the state of a crawl manager.
        # a dict job key : (spider, job_args_override)
        self._running_job_keys: Dict[JobKey, Tuple[SpiderName, JobParams]] = {}

    @property
    def description(self):
        return self.__doc__

    def add_argparser_options(self):
        super().add_argparser_options()
        if self.spider is None:
            self.argparser.add_argument("spider", help="Spider name")
        self.argparser.add_argument("--spider-args", help="Spider arguments dict in json format", default="{}")
        self.argparser.add_argument("--job-settings", help="Job settings dict in json format", default="{}")
        self.argparser.add_argument("--units", help="Set default number of ScrapyCloud units for each job", type=int)

    def parse_args(self) -> Namespace:
        args = super().parse_args()
        if self.spider is None:
            self.spider = args.spider
        return args

    def get_job_settings(self, override: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        job_settings = json.loads(self.args.job_settings)
        if override:
            job_settings.update(override)
        return job_settings

    def schedule_spider_with_jobargs(
        self,
        job_args_override: Optional[JobParams] = None,
        spider: Optional[SpiderName] = None,
    ) -> Optional[JobKey]:
        job_args_override = job_args_override or {}
        spider = spider or self.spider
        if spider is not None:
            schedule_args: ScheduleArgs = json.loads(self.args.spider_args)
            spider_args = job_args_override.get("spider_args") or {}
            schedule_args.update(job_args_override)
            schedule_args.pop("spider_args", None)
            schedule_args.update(spider_args)

            job_settings_override = schedule_args.get("job_settings", None)
            schedule_args["job_settings"] = self.get_job_settings(job_settings_override)
            schedule_args["units"] = schedule_args.get("units", self.args.units)
            jobkey = self.schedule_spider(spider, **schedule_args)
            if jobkey is not None:
                self._running_job_keys[jobkey] = spider, job_args_override
                self.stats.inc_value("crawlmanager/scheduled_jobs")
            return jobkey
        return None

    def check_running_jobs(self) -> Dict[JobKey, Outcome]:
        outcomes = {}
        running_job_keys = sorted(self._running_job_keys, reverse=self.running_jobs_reverse_check)
        while running_job_keys:
            jobkey = running_job_keys.pop(0)
            if (outcome := self.is_finished(jobkey)) is not None:

                metadata = self.get_job_metadata(jobkey)
                self.finished_metadata_hook(jobkey, metadata)

                spider, job_args_override = self._running_job_keys.pop(jobkey)
                if outcome in self.failed_outcomes:
                    _LOG.warning(f"Job {jobkey} finished with outcome {outcome}.")
                    if job_args_override is not None:
                        job_args_override = job_args_override.copy()
                    self.bad_outcome_hook(spider, outcome, job_args_override, jobkey)
                else:
                    self.finished_ok_hook(spider, outcome, job_args_override, jobkey)
                outcomes[jobkey] = outcome

                scrapystats = self.get_metadata_key(metadata, "scrapystats")
                self.aggregate_spider_stats(JobDict(spider=spider, key=jobkey, scrapystats=scrapystats))
            else:
                self.running_job_hook(jobkey)
        _LOG.info(f"There are {len(self._running_job_keys)} jobs still running.")

        return outcomes

    def running_job_hook(self, jobkey: JobKey):
        """
        Allows to add some reaction on each running job. Here we only pass jobkey, as it is more
        economical than getting metadata, and not needed unless you want to implement the hook.
        Just use in your implementation, something like
        metadata = self.get_job_metadata(jobkey)
        """

    def finished_metadata_hook(self, jobkey: JobKey, metadata: JobMeta):
        """
        allow to add some reaction on each finished job, based solely on its metadata.
        Use self.get_metadata_key(metadata, <key>) in order to get metadata with handled retries.
        """

    def bad_outcome_hook(self, spider: SpiderName, outcome: Outcome, job_args_override: JobParams, jobkey: JobKey):
        if self.get_close_reason() is None:
            self.set_close_reason(outcome)

    def finished_ok_hook(self, spider: SpiderName, outcome: Outcome, job_args_override: JobParams, jobkey: JobKey):
        pass

    def workflow_loop(self) -> bool:
        outcomes = self.check_running_jobs()
        if outcomes:
            return False
        if not self._running_job_keys:
            self.schedule_spider_with_jobargs()
        return True

    def resume_running_job_hook(self, job: JobDict):
        key = job["key"]
        spider_args = job.get("spider_args", {}).copy()
        job_args_override = JobParams(
            {
                "tags": job["tags"],
                "spider_args": spider_args,
            }
        )
        self._running_job_keys[key] = job["spider"], job_args_override
        _LOG.info(f"added running job {key}")

    def on_close(self):
        job = self.get_job()
        if job:
            close_reason = self.get_close_reason()
            self.finish(close_reason=close_reason)


class PeriodicCrawlManager(CrawlManager):
    """
    Schedule a spider periodically, waiting for the previous job to finish before scheduling it again with same
    parameters. Don't forget to set loop mode.
    """

    def bad_outcome_hook(self, spider: str, outcome: Outcome, job_args_override: JobParams, jobkey: JobKey):
        pass

    def workflow_loop(self) -> bool:
        self.check_running_jobs()
        if not self._running_job_keys:
            self.schedule_spider_with_jobargs()
        return True

    def on_close(self):
        pass


class GeneratorCrawlManagerProtocol(CrawlManagerProtocol, Protocol):

    _jobuids: DupesFilterProtocol
    spider: Optional[SpiderName]

    @abc.abstractmethod
    def _workflow_step_gen(self, max_next_params: int) -> Generator[Tuple[str, Optional[JobKey]], None, None]:
        ...

    @abc.abstractmethod
    def get_max_next_params(self) -> int:
        ...


class GeneratorCrawlManager(CrawlManager, GeneratorCrawlManagerProtocol):
    """
    Schedule a spider periodically, each time with different parameters yielded by a generator, until stop.
    Number of simultaneos spider jobs will be limited by max running jobs (see WorkFlowManager).
    Instructions:
    - Override set_parameters_gen() method. It must implement a generator of dictionaries, each one being
      the spider arguments (argument name: argument value) passed to the spider on each successive job.
    - Don't forget to set loop mode and max jobs (which defaults to infinite).
    """

    MAX_RETRIES = 0

    def __init__(self):
        super().__init__()
        self.__parameters_gen: Generator[ScheduleArgs, None, None] = self.set_parameters_gen()
        self.__additional_jobs: List[FullJobParams] = []
        self.__delayed_jobs: List[FullJobParams] = []
        self.__next_job_seq = 1
        self._jobuids = self.create_dupe_filter()

    def get_delayed_jobs(self) -> List[FullJobParams]:
        return deepcopy(self.__delayed_jobs)

    def add_delayed_jobs(self, params: FullJobParams):
        self.__delayed_jobs.append(params)

    @classmethod
    def create_dupe_filter(cls) -> DupesFilterProtocol:
        return BloomFilter(max_elements=1e6, error_rate=1e-6)

    def spider_running_count(self, spider: SpiderName) -> int:
        count = 0
        for spidername, _ in self._running_job_keys.values():
            if spidername == spider:
                count += 1
        return count

    def bad_outcome_hook(self, spider: SpiderName, outcome: Outcome, job_args_override: JobParams, jobkey: JobKey):
        if outcome == "cancelled":
            return
        if self.MAX_RETRIES == 0:
            return
        spider_args = job_args_override.setdefault("spider_args", {})
        tags = job_args_override["tags"]
        retries = get_jobseq(tags)[1]
        if retries < self.MAX_RETRIES:
            for t in list(tags):
                if t.startswith("RETRIED_FROM="):
                    tags.remove(t)
                    break
            tags.append(f"RETRIED_FROM={jobkey}")
            spider_args["retry_num"] = str(retries + 1)
            try:
                retry_override = self.get_retry_override(spider, outcome, job_args_override, jobkey)
                for rtag in retry_override.pop("tags", []):
                    if rtag not in tags:
                        tags.append(rtag)
                job_args_override.setdefault("spider_args", {}).update(retry_override.pop("spider_args", {}))
                job_args_override.setdefault("job_settings", {}).update(retry_override.pop("job_settings", {}))
                job_args_override.update(retry_override)
            except StopRetry as e:
                _LOG.info(f"Job {jobkey} failed with reason '{outcome}'. Will not be retried. Reason: {e}")
            else:
                self.add_job(spider, job_args_override)
                _LOG.info(
                    f"Job {jobkey} failed with reason '{outcome}'. Retrying ({retries + 1} of {self.MAX_RETRIES})."
                )

    def get_retry_override(
        self, spider: SpiderName, outcome: Outcome, job_args_override: JobParams, jobkey: JobKey
    ) -> JobParams:
        return JobParams({})

    def add_job(self, spider: SpiderName, job_args_override: JobParams):
        params = cast(FullJobParams, (job_args_override or {}).copy())
        params["spider"] = spider
        self.__additional_jobs.append(params)

    @abc.abstractmethod
    def set_parameters_gen(self) -> Generator[ScheduleArgs, None, None]:
        ...

    def __add_jobseq_tag(self, params: FullJobParams):
        tags = params.setdefault("tags", [])
        jobseq_tag = None
        for tag in tags:
            if tag.startswith("JOBSEQ="):
                jobseq_tag = tag
                break
        if jobseq_tag is None:
            jobseq_tag = f"JOBSEQ={self.__next_job_seq:010d}"
            self.__next_job_seq += 1
        else:
            tags.remove(jobseq_tag)
            jobseq, repn = get_jobseq([jobseq_tag])
            jobseq_tag = f"JOBSEQ={jobseq:010d}.r{repn + 1}"
        tags.append(jobseq_tag)

    def get_running_spiders(self) -> Set[SpiderName]:
        return set(sp for sp, _ in self._running_job_keys.values())

    def get_delayed_spiders(self) -> Set[SpiderName]:
        return set(np["spider"] for np in self.__delayed_jobs)

    def spider_delayed_count(self, spider: SpiderName) -> int:
        count = 0
        for np in self.__delayed_jobs:
            if np["spider"] == spider:
                count += 1
        return count

    def can_schedule_job_with_params(self, params: FullJobParams) -> bool:
        """
        If returns False, delay the scheduling of the params.
        """
        return True

    def _fulljobparams_from_spiderargs(self, schedule_args: ScheduleArgs) -> FullJobParams:
        spider = schedule_args.get("spider", self.spider)
        spider_args = deepcopy(schedule_args)
        for key in tuple(FullJobParams.__annotations__):
            spider_args.pop(key, None)
        result = FullJobParams(
            {
                "spider": spider,
                "spider_args": spider_args,
            }
        )
        if "units" in schedule_args:
            result["units"] = schedule_args["units"]
        if "tags" in schedule_args:
            result["tags"] = schedule_args["tags"]
        if "job_settings" in schedule_args:
            result["job_settings"] = schedule_args["job_settings"]
        if "project_id" in schedule_args:
            result["project_id"] = schedule_args["project_id"]
        return result

    def _workflow_step_gen(self, max_next_params: int) -> Generator[Tuple[str, Optional[JobKey]], None, None]:
        new_params: List[FullJobParams] = []
        next_params: Optional[FullJobParams]

        while len(new_params) < max_next_params:
            next_params = None
            for idx, np in enumerate(self.__delayed_jobs):
                if self.can_schedule_job_with_params(np):
                    next_params = np
                    self.__delayed_jobs = self.__delayed_jobs[:idx] + self.__delayed_jobs[idx + 1:]
                    break
            else:
                for idx, np in enumerate(self.__additional_jobs):
                    if self.can_schedule_job_with_params(np):
                        next_params = np
                        self.__additional_jobs = self.__additional_jobs[:idx] + self.__additional_jobs[idx + 1:]
                        break
                else:
                    try:
                        np = self._fulljobparams_from_spiderargs(next(self.__parameters_gen))
                    except StopIteration:
                        break
                    else:
                        spider = np.get("spider", self.spider)
                        assert spider, f"No spider set for parameters {np}"
                        np["spider"] = spider
                        if self.can_schedule_job_with_params(np):
                            next_params = np
                        else:
                            self.__delayed_jobs.append(np)
                            continue

            if next_params is None:
                break
            spider_args = next_params.get("spider_args") or {}
            jobuid = self.get_job_unique_id({"spider": next_params["spider"], "spider_args": spider_args})
            if jobuid in self._jobuids:
                _LOG.warning(f"Job with parameters {next_params} was already scheduled. Skipped.")
                continue
            new_params.append(next_params)

        for next_params in new_params:
            spider = next_params.pop("spider")
            self.__add_jobseq_tag(next_params)
            yield jobuid, self.schedule_spider_with_jobargs(next_params, spider)

    def get_max_next_params(self) -> int:
        return self.max_running_jobs - len(self._running_job_keys)

    def workflow_loop(self) -> bool:
        self.check_running_jobs()
        max_next_params = self.get_max_next_params()
        retval = False
        for jobuid, jobid in self._workflow_step_gen(max_next_params):
            retval = True
            if jobid is not None:
                self._jobuids.add(jobuid)
        return retval or bool(self._running_job_keys)

    @staticmethod
    def get_job_unique_id(job: JobDict) -> str:
        jdict = job.get("spider_args", {}).copy()
        jdict["spider"] = job["spider"]
        for k, v in jdict.items():
            jdict[k] = str(v)
        jid = json.dumps(jdict, sort_keys=True)
        return hashstr(jid)

    def resume_running_job_hook(self, job: JobDict):
        super().resume_running_job_hook(job)
        jobuid = self.get_job_unique_id(job)
        self._jobuids.add(jobuid)
        self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)

    def resume_finished_job_hook(self, job: JobDict):
        jobuid = self.get_job_unique_id(job)
        if jobuid not in self._jobuids:
            self._jobuids.add(jobuid)
        self.__next_job_seq = max(self.__next_job_seq, get_jobseq(job["tags"])[0] + 1)

    def resume_workflow(self):
        super().resume_workflow()
        _LOG.info(f"Next job sequence number: {self.__next_job_seq}")

    def on_close(self):
        self._jobuids.close()


class AsyncSchedulerCrawlManagerMixin(BaseLoopScriptAsyncMixin, GeneratorCrawlManagerProtocol):
    async def _async_workflow_step_gen(self, max_next_params: int) -> AsyncGenerator[Tuple[str, JobKey], None]:
        jobuids_cors = list(self._workflow_step_gen(max_next_params))
        if jobuids_cors:
            jobuids, cors = zip(*jobuids_cors)
            results: List[JobKey] = await asyncio.gather(*cors)
            for jobuid, jobid in zip(jobuids, results):
                yield jobuid, jobid

    async def workflow_loop(self) -> bool:  # type: ignore
        self.check_running_jobs()
        max_next_params = self.get_max_next_params()
        retval = False
        async for jobuid, jobid in self._async_workflow_step_gen(max_next_params):
            retval = True
            if jobid is not None:
                self._jobuids.add(jobuid)
        return retval or bool(self._running_job_keys)

    async def schedule_spider_with_jobargs(  # type: ignore
        self,
        job_args_override: Optional[JobParams] = None,
        spider: Optional[SpiderName] = None,
    ) -> Optional[JobKey]:
        job_args_override = job_args_override or {}
        spider = spider or self.spider
        assert spider is not None
        schedule_args: ScheduleArgs = json.loads(self.args.spider_args)
        spider_args = job_args_override.get("spider_args") or {}
        schedule_args.update(job_args_override)
        schedule_args.pop("spider_args", None)
        schedule_args.update(spider_args)

        job_settings_override = schedule_args.get("job_settings", None)
        schedule_args["job_settings"] = self.get_job_settings(job_settings_override)
        schedule_args["units"] = schedule_args.get("units", self.args.units)
        jobkey = await self.async_schedule_spider(spider, **schedule_args)
        if jobkey is not None:
            self._running_job_keys[jobkey] = spider, job_args_override
            self.stats.inc_value("crawlmanager/scheduled_jobs")
        return jobkey
